(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(*  This example is based on an example designed for OCamlP3l.         *)
(*                                                                     *)
(***********************************************************************)

(* $Id: rdec2.mli,v 1.13 2014/01/21 09:21:18 fclement Exp $ *)

(*****************************************************************)
(* Module for the representation of 2D regular rectangular grids *)
(* with regular decomposition capabilities.                      *)
(* Overlapping of the decomposition is parameterized.            *)
(* Uses scanning buffers, which needs at least OCaml-3.08.4.     *)
(*****************************************************************)

(****************************************************************************)
(* A config file contains a regular_decomposition record, ie:               *)
(*   a computational_rectangular_grid record:                               *)
(*     a rectangular_grid record:                                           *)
(*       a rectangle record:                                                *)
(*         x0 and y0, coordinates of the lower-left corner of the rectangle *)
(*         lx and ly, length of the rectangle in both directions            *)
(*       nx and ny, number of intervals in both directions                  *)
(*     px and py, half the overlapping of the subdomains in both directions *)
(*   nsbdx and nsbdy, number of subdomains in both directions.              *)
(*                                                                          *)
(* A valid regular_decomposition must satisfy:                              *)
(*   lx > 0. and ly > 0.                                                    *)
(*   nx > 0 and ny > 0                                                      *)
(*   px > 0 and py > 0                                                      *)
(*   px <= nx / nsbdx and py <= ny / nsbdy                                  *)
(*   nsbdx > 0 and nsbdy > 0                                                *)
(*   nsbdx <= nx and nsbdy <= ny                                            *)
(*   nx mod nsbdx = 0 and ny mod nsbdy = 0                                  *)
(****************************************************************************)

type rectangle
and rectangular_grid
and computational_rectangular_grid
and regular_decomposition
;;

type boundary and boundary_size;;
type subdomain = regular_decomposition * int;;
type interface = boundary array;;

exception Wrong_Input of string;;
exception Out_of_range of int;;

val number_of_subdomains_of_regular_decomposition :
    regular_decomposition -> int * int
;;
val print_computational_rectangular_grid :
    out_channel -> computational_rectangular_grid -> unit
;;
val parse_rdec_config_file : string -> regular_decomposition;;
val check_regular_decomposition : regular_decomposition -> unit;;
val neighbors_of : subdomain -> int list;;
val computational_rectangular_grid_of_subdomain :
    subdomain -> computational_rectangular_grid
;;
val proj_boundary : subdomain -> boundary -> (int * boundary) -> boundary;;
val normalize_boundary : boundary -> boundary;;
val print_boundary : out_channel -> boundary -> unit;;
val read_boundary : Scanf.Scanning.scanbuf -> boundary_size -> boundary;;
val size_of_boundary : boundary -> boundary_size;;
val make_zero_boundary : boundary_size -> boundary;;
val make_zero_interface : regular_decomposition -> interface;;
val for_all_subdomains : int * int -> (int -> unit) -> unit;;
