(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: helper.ml,v 1.5 2014/01/21 09:21:18 fclement Exp $ *)

let scalar_prod v1 v2 =
  let prod = ref 0 in
  for i = 0 to Array.length v1 - 1 do
    prod := !prod + v1.(i) * v2.(i)
  done; !prod
;;
(* OCaml scalar product *)

module IO : sig
  val input_vector : in_channel -> (int array) option
  (* Input a vector given an [in_channel] *)
  val input_matrix : in_channel -> ((int array) array) option
  (* Input a matrix given an [in_channel] *)
  val input_fvector : in_channel -> (float array) option
  (* Input a vector of [float]s on a given [in_channel] *)
  val input_tridiag : in_channel -> Tridiag.t option
  (* Input a tridiagonal matrix given an [in_channel] *)
  val output_vector : out_channel -> int array -> unit;;
  (* Output a vector on a given [out_channel] *)
  val output_raw : out_channel -> (float array) array -> unit;;
  (* Dump an array of arrays in a row fashion *)
end = struct
  (* Input output operations *)
  open Parser
  open Parser_char

  let parse_vec p =
    pspaced (pchar '[') >>
      (pspaced (sep_by1 (pspaced (pchar_ign ';')) (pspaced p))) >>= fun l ->
        pspaced (pchar ']') >> return (Array.of_list l)

  let parse_vector = parse_vec pint
  let parse_matrix = parse_vec parse_vector
  let parse_fvector = parse_vec pfloat
  let parse_tridiag =
    parse_fvector >>= fun d0 ->
    parse_fvector >>= fun d1 ->
    parse_fvector >>= fun d2 ->
    return (Tridiag.create d0 d1 d2)

  let slurp ic =
    let rec slurp lines =
      match
        try Some (input_line ic)
        with End_of_file -> None
      with
      | Some l -> slurp (l :: lines)
      | None -> String.concat "\n" (List.rev lines) in
    slurp []

  let sanitize matopt =
    match matopt with
    | Some mat ->
      let len = Array.length mat in
      let rec sanitize i l =
        if i >= len then true else
        Array.length mat.(i) = l && sanitize (succ i) l in
      if not (sanitize 1 (Array.length mat.(0))) then None else matopt
    | None -> None

  let input_vector ic = parse parse_vector (slurp ic)
  let input_fvector ic = parse parse_fvector (slurp ic)
  let input_matrix ic = sanitize (parse parse_matrix (slurp ic))
  let input_tridiag ic = parse parse_tridiag (slurp ic)

  open Format

  let output_vector oc vec =
    let len = Array.length vec in
    let ppf = formatter_of_out_channel oc in
    let rec output_vector ppf i =
      if i >= len then () else
      if i >= len - 1 then
        fprintf ppf "%d" vec.(i)
      else
        fprintf ppf "%d;@ %a" vec.(i) output_vector (succ i) in
    fprintf ppf "@[<hv 2>[@ %a@ ]@]@." output_vector 0

  let output_raw oc vecs =
    let output_vec v =
      Array.iter (Printf.fprintf oc "%f\n") v in
    Array.iter output_vec vecs; flush oc
end;;

let fetch_std_data mat_file vec_file =
  let parse_error f =
    Printf.eprintf "Parsing error while processing %S.\n" f in
  let (mat_chan, vec_chan) = (open_in mat_file, open_in vec_file) in
  match IO.input_matrix mat_chan, IO.input_vector vec_chan with
  | Some mat, Some vec -> (mat, vec)
  | Some _, None -> parse_error mat_file; exit 2
  | None, Some _ -> parse_error vec_file; exit 2
  | None, None -> parse_error mat_file; parse_error vec_file; exit 2
;;

let fetch_tri_data mat_file vec_file =
  let parse_error f =
    Printf.eprintf "Parsing error while processing %S.\n" f in
  let (mat_chan, vec_chan) = (open_in mat_file, open_in vec_file) in
  match IO.input_tridiag mat_chan, IO.input_fvector vec_chan with
  | Some mat, Some vec -> (mat, vec)
  | Some _, None -> parse_error mat_file; exit 2
  | None, Some _ -> parse_error vec_file; exit 2
  | None, None -> parse_error mat_file; parse_error vec_file; exit 2
;;
