(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: sk.ml,v 1.28 2014/01/21 09:21:15 fclement Exp $ *)

type ('a, 'b) sum = ('a, 'b) Sklstream.sum = Inl of 'a | Inr of 'b;;

module Eq : sig
  type ('a, 'b) t
  val refl : unit -> ('a, 'a) t
  val app : ('a, 'b) t -> 'a -> 'b
end = struct
  type ('a, 'b) t = 'a -> 'b
  external id : 'a -> 'a = "%identity"
  let refl () = id
  let app e = fun x -> e x
end
;;

module rec Skl : sig
  module type ATOM = sig
    type a and b
    type x

    val color : int
    val params : x
    val func : (x, a, b) Expor.llexpr
  end
  module type PIPE = sig
    type a and b
    type c

    val sk1 : (a, c) Skl.t
    val sk2 : (c, b) Skl.t
  end
  module type PROD = sig
    type a and b
    type a0 and a1 and b0 and b1

    val eq_in : (a, (a0 * a1)) Eq.t
    val eq_out : ((b0 * b1), b) Eq.t
    val sk1 : (a0, b0) Skl.t
    val sk2 : (a1, b1) Skl.t
  end
  module type SUM = sig
    type a and b
    type aa and ab

    val eq_in : (a, (aa, ab) sum) Eq.t
    val sk1 : (aa, b) Skl.t
    val sk2 : (ab, b) Skl.t
  end

  type ('a, 'b) atom = (module Skl.ATOM with type a = 'a and type b = 'b)
  type ('a, 'b) id = ('a, 'b) Eq.t
  type ('a, 'b) prod = (module Skl.PROD with type a = 'a and type b = 'b)
  type ('a, 'b) pipe = (module Skl.PIPE with type a = 'a and type b = 'b)
  type ('a, 'b) rails = (('a, 'b) Skl.t) array
  type ('a, 'b) sum = (module Skl.SUM with type a = 'a and type b = 'b)
  type ('a, 'b) par = (('a, 'b) Skl.t * ('a, 'b) Skl.t)
  type ('a, 'b) loop = (('a, 'b) Eq.t * ('a, bool) Skl.t * ('a, 'a) Skl.t)

  type ('a, 'b) t =
    | Atom of ('a, 'b) atom
    | Id of ('a, 'b) id
    | Prod of ('a, 'b) prod
    | Pipe of ('a, 'b) pipe
    | Rails of ('a, 'b) rails
    | Sum of ('a, 'b) sum
    | Par  of ('a, 'b) par
    | Loop of ('a, 'b) loop
end = Skl
;;

type ('a, 'b, 'd) atom_runner =
  { f_atom : 'x. 'x -> ('x, 'a, 'b) Expor.llexpr -> int -> 'd; }
;;
type ('a, 'b, 'd) id_runner =
  { f_id : ('a -> 'b) -> 'd; }
;;
type ('a, 'b, 'd) prod_runner =
  { f_prod : 'a0 'b0 'a1 'b1.
      ('a -> ('a0 * 'a1)) -> (('b0 * 'b1) -> 'b) ->
      ('a0, 'b0) Skl.t -> ('a1, 'b1) Skl.t -> 'd; }
;;
type ('a, 'c, 'd) pipe_runner =
  { f_pipe : 'b. ('a, 'b) Skl.t -> ('b, 'c) Skl.t -> 'd; }
;;
type ('a, 'b, 'd) par_runner =
  { f_par : ('a, 'b) Skl.t -> ('a, 'b) Skl.t -> 'd; }
;;
type ('a, 'b, 'd) rails_runner =
  { f_rails : (('a, 'b) Skl.t) array -> 'd; }
;;
type ('a, 'b, 'd) sum_runner =
  { f_sum : 'aa 'ab.
      ('a -> ('aa, 'ab) sum) -> ('aa, 'b) Skl.t -> ('ab, 'b) Skl.t -> 'd; }
;;
type ('a, 'b, 'd) loop_runner =
  { f_loop : ('a -> 'b) -> ('a, bool) Skl.t -> ('a, 'a) Skl.t -> 'd; }
;;

let with_atom (type s) (type t) atom runner =
  let module Atom = (val atom : Skl.ATOM with type a = s and type b = t) in
  runner.f_atom Atom.params Atom.func Atom.color
;;

let with_id id runner =
  runner.f_id (Eq.app id)
;;

let with_prod (type s) (type t) prod runner =
  let module Prod = (val prod : Skl.PROD with type a = s and type b = t) in
  runner.f_prod (Eq.app Prod.eq_in) (Eq.app Prod.eq_out) Prod.sk1 Prod.sk2
;;

let with_pipe (type s) (type t) pipe runner =
  let module Pipe = (val pipe : Skl.PIPE with type a = s and type b = t) in
  runner.f_pipe Pipe.sk1 Pipe.sk2
;;

let with_par (sk1, sk2) runner =
  runner.f_par sk1 sk2
;;

let with_rails ska runner =
  runner.f_rails ska
;;

let with_sum (type s) (type t) sum runner =
  let module Sum = (val sum : Skl.SUM with type a = s and type b = t) in
  runner.f_sum (Eq.app Sum.eq_in) Sum.sk1 Sum.sk2
;;

let with_loop (e, p, sk) runner =
  runner.f_loop (Eq.app e) p sk
;;

let c_atom (type s) (type t) (type u) params func color =
  let module Atom = struct
    type a = s and b = t and x = u

    let color = color
    let params = params
    let func = func
  end in
  Skl.Atom (module Atom : Skl.ATOM with type a = s and type b = t)
;;

let c_id () =
  Skl.Id (Eq.refl ())
;;

let c_prod (type s) (type t) (type u) (type v) sk1 sk2 =
  let module Prod = struct
    type a0 = s and a1 = t and b0 = u and b1 = v
    type a = (a0 * a1) and  b = (b0 * b1)

    let sk1 = sk1 and sk2 = sk2
    let eq_in = Eq.refl () and eq_out = Eq.refl ()
  end in
  Skl.Prod (module Prod : Skl.PROD with type a = (s * t) and type b = (u * v))
;;

let c_pipe (type s) (type t) (type u) sk1 sk2 =
  let module Pipe = struct
    type a = s and b = t and c = u

    let sk1 = sk1
    let sk2 = sk2
  end in
  Skl.Pipe (module Pipe : Skl.PIPE with type a = s and type b = t)
;;

let c_par sk1 sk2 =
  Skl.Par (sk1, sk2)
;;

let c_rails ska =
  Skl.Rails ska
;;

let c_sum (type s) (type t) (type u) sk1 sk2 =
  let module Sum = struct
    type a = (s, t) sum and b = u
    type aa = s and ab = t

    let eq_in = Eq.refl ()
    let sk1 = sk1 and sk2 = sk2
  end in
  Skl.Sum (module Sum : Skl.SUM with type a = (s, t) sum and type b = u)
;;

let c_loop p sk =
  Skl.Loop (Eq.refl (), p, sk)
;;

let colorize (type s) (type t) atom color =
  let module Atom = (val atom : Skl.ATOM with type a = s and type b = t) in
  c_atom Atom.params Atom.func color
;;
